@@ -5,7 +5,7 @@
     Flask-Security decorators module
 
     :copyright: (c) 2012-2019 by Matt Wright.
-    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
@@ -14,7 +14,7 @@ import datetime
 from functools import wraps
 import typing as t
 
-from flask import Response, _request_ctx_stack, abort, current_app, g, redirect, request
+from flask import Response, abort, current_app, g, redirect, request
 from flask_login import current_user, login_required  # noqa: F401
 from flask_principal import Identity, Permission, RoleNeed, identity_changed
 from flask_wtf.csrf import CSRFError
@@ -26,12 +26,13 @@ from .utils import (
     FsPermNeed,
     config_value,
     do_flash,
-    find_user,
     get_message,
     get_url,
+    lookup_identity,
     check_and_update_authn_fresh,
     json_error_response,
     set_request_attr,
+    url_for_security,
 )
 
 # Convenient references
@@ -97,21 +98,21 @@ def default_reauthn_handler(within, grace):
     m, c = get_message("REAUTHENTICATION_REQUIRED")
 
     if _security._want_json(request):
+        from .webauthn import has_webauthn
+
         is_us = config_value("UNIFIED_SIGNIN")
         payload = json_error_response(errors=m)
         payload["reauth_required"] = True
         payload["unified_signin_enabled"] = is_us
+        payload["has_webauthn_verify_credential"] = has_webauthn(
+            current_user, config_value("WAN_ALLOW_AS_VERIFY")
+        )
         return _security._render_json(payload, 401, None, None)
 
-    if config_value("UNIFIED_SIGNIN"):
-        view = _security.us_verify_url
-    else:
-        view = _security.verify_url
-    if view:
-        do_flash(m, c)
-        redirect_url = get_url(view, qparams={"next": request.url})
-        return redirect(redirect_url)
-    abort(401)
+    view = "us_verify" if config_value("UNIFIED_SIGNIN") else "verify"
+    do_flash(m, c)
+    redirect_url = url_for_security(view, next=request.url)
+    return redirect(redirect_url)
 
 
 def default_unauthz_handler(func, params):
@@ -153,7 +154,6 @@ def _check_token():
 
     if user and user.is_authenticated:
         app = current_app._get_current_object()
-        _request_ctx_stack.top.user = user
         identity_changed.send(app, identity=Identity(user.fs_uniquifier))
         return True
 
@@ -164,14 +164,14 @@ def _check_http_auth():
     auth = request.authorization or BasicAuth(username=None, password=None)
     if not auth.username:
         return False
-    user = find_user(auth.username)
+    user = lookup_identity(auth.username)
     if user and not user.active:
         return False
 
     if user and user.verify_and_update_password(auth.password):
         _security.datastore.commit()
+        _security.login_manager._update_request_context_with_user(user)
         app = current_app._get_current_object()
-        _request_ctx_stack.top.user = user
         identity_changed.send(app, identity=Identity(user.fs_uniquifier))
         return True
 
@@ -212,7 +212,7 @@ def handle_csrf(method: t.Optional[str]) -> None:
         if method in config_value("CSRF_PROTECT_MECHANISMS"):
             _csrf.protect()  # type: ignore
         else:
-            _request_ctx_stack.top.fs_ignore_csrf = True
+            set_request_attr("fs_ignore_csrf", True)
 
 
 def http_auth_required(realm: t.Any) -> DecoratedView:
@@ -238,7 +238,7 @@ def http_auth_required(realm: t.Any) -> DecoratedView:
             else:
                 r = _security.default_http_auth_realm if callable(realm) else realm
                 h = {"WWW-Authenticate": 'Basic realm="%s"' % r}
-                return _security._unauthn_handler(["basic"], headers=h)
+                return _security._unauthn_handler(["basic"], h)
 
         return wrapper
 
@@ -277,7 +277,7 @@ def auth_required(
     grace: t.Optional[t.Union[int, float, t.Callable[[], datetime.timedelta]]] = None
 ) -> DecoratedView:
     """
-    Decorator that protects endpoints through multiple mechanisms
+    Decorator that protects endpoints through multiple mechanisms.
     Example::
 
         @app.route('/dashboard')
@@ -443,7 +443,7 @@ def unauth_csrf(
                 config_value("CSRF_IGNORE_UNAUTH_ENDPOINTS")
                 and not current_user.is_authenticated
             ):
-                _request_ctx_stack.top.fs_ignore_csrf = True
+                set_request_attr("fs_ignore_csrf", True)
             else:
                 try:
                     _csrf.protect()
