@@ -5,7 +5,7 @@
     Flask-Security utils module
 
     :copyright: (c) 2012-2019 by Matt Wright.
-    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 import abc
@@ -17,13 +17,12 @@ import hmac
 from pkg_resources import parse_version
 import time
 import typing as t
-import warnings
 from urllib.parse import parse_qsl, parse_qs, urlsplit, urlunsplit, urlencode
 import urllib.request
 import urllib.error
+import warnings
 
 from flask import (
-    _request_ctx_stack,
     after_this_request,
     current_app,
     flash,
@@ -33,7 +32,6 @@ from flask import (
     session,
     url_for,
 )
-from flask.json import JSONEncoder
 from flask_login import login_user as _login_user
 from flask_login import logout_user as _logout_user
 from flask_login import current_user
@@ -52,6 +50,8 @@ from .signals import user_authenticated
 
 if t.TYPE_CHECKING:  # pragma: no cover
     from flask import Flask, Response
+    from flask.typing import ResponseValue
+    from flask_wtf import FlaskForm
     from .datastore import User
 
 SB = t.Union[str, bytes]
@@ -71,7 +71,7 @@ def _(translate):
 def get_request_attr(name: str) -> t.Any:
     """Retrieve a request local attribute.
 
-    Currently public attributes are:
+    Current public attributes are:
 
     **fs_authn_via**
         will be set to the authentication mechanism (session, token, basic)
@@ -80,12 +80,14 @@ def get_request_attr(name: str) -> t.Any:
     Returns None if attribute doesn't exist.
 
     .. versionadded:: 4.0.0
+    .. versionchanged:: 4.1.5
+        Use 'g' rather than request_ctx stack which is going away post Flask 2.2
     """
-    return getattr(_request_ctx_stack.top, name, None)
+    return getattr(g, name, None)
 
 
 def set_request_attr(name, value):
-    return setattr(_request_ctx_stack.top, name, value)
+    return setattr(g, name, value)
 
 
 """
@@ -98,7 +100,6 @@ if get_quart_status():  # pragma: no cover
         _datastore.commit()
         return response
 
-
 else:
 
     def view_commit(response=None):
@@ -262,7 +263,7 @@ def check_and_update_authn_fresh(
     fs_gexp = session.get("fs_gexp", None)
     if fs_gexp:
         if now.timestamp() < fs_gexp:
-            # Within grace period - extend it and we're good.
+            # Within grace period - extend it, and we're good.
             session["fs_gexp"] = grace_ts
             return True
 
@@ -273,7 +274,7 @@ def check_and_update_authn_fresh(
 
     authn_time = datetime.datetime.utcfromtimestamp(session["fs_paa"])
     # allow for some time drift where it's possible authn_time is in the future
-    # but lets be cautious and not allow arbitrary future times
+    # but let's be cautious and not allow arbitrary future times
     delta = now - authn_time
     if within > delta > -within:
         session["fs_gexp"] = grace_ts
@@ -347,22 +348,6 @@ def verify_and_update_password(password: SB, user: "User") -> bool:
     return verified
 
 
-def encrypt_password(password):  # pragma: no cover
-    """Encrypt the specified plaintext password.
-
-    It uses the configured encryption options.
-
-    .. deprecated:: 2.0.2
-       Use :func:`hash_password` instead.
-
-    :param password: The plaintext password to encrypt
-    """
-    warnings.warn(
-        "Please use hash_password instead of encrypt_password.", DeprecationWarning
-    )
-    return hash_password(password)
-
-
 def hash_password(password: SB) -> t.Any:
     """Hash the specified plaintext password.
 
@@ -513,7 +498,8 @@ def url_for_security(endpoint: str, **values: t.Union[str, bool]) -> str:
     :param _method: if provided this explicitly specifies an HTTP method.
     """
     endpoint = get_security_endpoint_name(endpoint)
-    return url_for(endpoint, **values)
+    # mypy is complaining about this - but I think it's wrong?
+    return url_for(endpoint, **values)  # type: ignore
 
 
 def validate_redirect_url(url: str) -> bool:
@@ -711,12 +697,6 @@ def send_mail(subject, recipient, template, **context):
     if isinstance(sender, LocalProxy):
         sender = sender._get_current_object()
 
-    # In Flask-Mail, sender can be a two element tuple -- (name, address)
-    if isinstance(sender, tuple) and len(sender) == 2:
-        sender = (str(sender[0]), str(sender[1]))
-    else:
-        sender = str(sender)
-
     _security._mail_util.send_mail(
         template, subject, recipient, sender, body, html, context.get("user", None)
     )
@@ -731,7 +711,13 @@ def get_token_status(token, serializer, max_age=None, return_data=False):
     :param max_age: The name of the max age config option. Can be one of
                     the following: ``CONFIRM_EMAIL``, ``LOGIN``,
                     ``RESET_PASSWORD``
+
+    .. deprecated:: 5.0.0
     """
+    warnings.warn(
+        "'get_token_status' is deprecated - use check_and_get_token_status instead",
+        DeprecationWarning,
+    )
     serializer = getattr(_security, serializer + "_serializer")
     max_age = get_max_age(max_age)
     user, data = None, None
@@ -764,7 +750,7 @@ def check_and_get_token_status(
     :param token: The token to check
     :param serializer_name: The name of the serializer. Can be one of the
                        following: ``confirm``, ``login``, ``reset``, ``us_setup``
-                       ``remember``, ``two_factor_validity``
+                       ``remember``, ``two_factor_validity``, ``wan``
     :param within: max age - passed as a timedelta
 
     :return: a tuple of (expired, invalid, data)
@@ -815,11 +801,12 @@ def get_identity_attribute(
     return {}
 
 
-def find_user(identity):
+def lookup_identity(identity):
     """
-    Validate identity - we go in order to figure out which user attribute the
-    request gave us. Note that we give up on the first 'match' even if that
-    doesn't yield a user. Why?
+    Lookup identity in DB.
+    This loops through, in order, SECURITY_USER_IDENTITY_ATTRIBUTES, and first
+    calls the mapper function to validate/normalize. Then the db.find_user is called
+    on the specified user model attribute.
     """
     for mapping in config_value("USER_IDENTITY_ATTRIBUTES"):
         attr = list(mapping.keys())[0]
@@ -958,23 +945,23 @@ def csrf_cookie_handler(response: "Response") -> "Response":
 
 
 def base_render_json(
-    form,
-    include_user=True,
-    include_auth_token=False,
-    additional=None,
-    error_status_code=400,
-):
+    form: "FlaskForm",
+    include_user: bool = True,
+    include_auth_token: bool = False,
+    additional: t.Optional[t.Dict[str, t.Any]] = None,
+    error_status_code: int = 400,
+) -> "ResponseValue":
     """
     This method is called by all views that return JSON responses.
     This fills in the response and then calls :meth:`.Security.render_json`
     which can be overridden by the app.
     """
-    has_errors = len(form.errors) > 0
-
     user = form.user if hasattr(form, "user") else None
-    if has_errors:
+    if form.errors:
         code = error_status_code
-        payload = json_error_response(errors=form.errors)
+        # wtforms 3.0 introduces form-level errors - these show up as part of the
+        # errors dict with a key of 'None'
+        payload = json_error_response(field_errors=form.errors)
     else:
         code = 200
         payload = dict()
@@ -1009,9 +996,18 @@ def base_render_json(
     return _security._render_json(payload, code, None, user)
 
 
+def simple_render_json(
+    additional: t.Optional[t.Dict[str, t.Any]] = None,
+) -> "ResponseValue":
+    payload = dict(csrf_token=csrf.generate_csrf())
+    if additional:
+        payload.update(additional)
+    return _security._render_json(payload, 200, None, None)
+
+
 def default_want_json(req):
     """Return True if response should be in json
-    N.B. do not call this directly - use security.want_json()
+    N.B. do not call this directly - use security._want_json()
 
     :param req: Flask/Werkzeug Request
     """
@@ -1028,37 +1024,40 @@ def default_want_json(req):
     return False
 
 
-def json_error_response(errors):
-    """Helper to create an error response that adheres to the openapi spec."""
-    if isinstance(errors, str):
-        # When the errors is a string, use the response/error/message format
-        response_json = dict(error=errors)
-    elif isinstance(errors, dict):
-        # When the errors is a dict, use the DefaultJsonErrorResponse
-        # (response/errors/name/messages) format
-        response_json = dict(errors=errors)
-    else:
-        raise TypeError("The errors argument should be either a str or dict.")
+def json_error_response(
+    errors: t.Optional[t.Union[str, list]] = None,
+    field_errors: t.Optional[t.Dict[str, list]] = None,
+) -> t.Dict[str, t.Any]:
+    """Helper to create an error response.
+
+    The "errors" key holds a simple list of errors - which is made up of any passed
+    errors (either a string or list) as well as the (localized) error msg from the
+    passed in form errors.
+
+    The "field_errors" key which is exactly what is returned from WTForms - namely
+    a dict of field-name: msg. For form-level errors (WTForms 3.0) the 'field-name' is
+    'None'
+    """
+    response_json: t.Dict[str, t.Union[list, t.Dict[str, list]]] = dict()
+    plain_errors = []
+    if errors:
+        if isinstance(errors, str):
+            plain_errors = [errors]
+        elif isinstance(errors, list):
+            plain_errors = errors
+        else:
+            raise TypeError("The errors argument should be either a str or list.")
+    if field_errors:
+        # This is default from WTForms - a dictionary of field name and list of errors
+        # we return that, as well as create a simple list of errors.
+        for e in field_errors.values():
+            plain_errors.extend(e)
+        response_json["field_errors"] = field_errors
+    response_json["errors"] = plain_errors
 
     return response_json
 
 
-class FsJsonEncoder(JSONEncoder):
-    """Flask-Security JSON encoder.
-    Extends Flask's JSONencoder to handle lazy-text.
-
-    .. versionadded:: 3.3.0
-    """
-
-    def default(self, obj):
-        from .babel import is_lazy_string
-
-        if is_lazy_string(obj):
-            return str(obj)
-        else:
-            return JSONEncoder.default(self, obj)
-
-
 def default_render_template(*args, **kwargs):
     return render_template(*args, **kwargs)
 
@@ -1170,8 +1169,7 @@ def password_complexity_validator(
             if kwargs:
                 user_info = list(kwargs.values())
         results = zxcvbn.zxcvbn(password, user_inputs=user_info)
-        if results["score"] > 2:
-            # Good or Strong
+        if results["score"] >= config_value("ZXCVBN_MINIMUM_SCORE"):
             return None
         # Should we return suggestions? Default forms don't really know what to do.
         if results["feedback"]["warning"]:
