@@ -5,7 +5,7 @@
     Flask-Security views module
 
     :copyright: (c) 2012 by Matt Wright.
-    :copyright: (c) 2019-2021 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 
     CSRF is tricky. By default all our forms have CSRF protection built in via
@@ -49,6 +49,7 @@ from .confirmable import (
     send_confirmation_instructions,
 )
 from .decorators import anonymous_user_required, auth_required, unauth_csrf
+from .forms import form_errors_munge
 from .passwordless import login_token_status, send_login_instructions
 from .proxies import _security, _datastore
 from .quart_compat import get_quart_status
@@ -66,22 +67,25 @@ from .recoverable import (
     send_reset_password_instructions,
     update_password,
 )
-from .registerable import register_user
+from .registerable import register_user, register_existing
+from .tf_plugin import (
+    create_recovery_codes,
+    tf_check_state,
+    tf_illegal_state,
+    tf_set_validity_token_cookie,
+)
 from .twofactor import (
     complete_two_factor_process,
-    generate_tf_validity_token,
-    is_tf_setup,
+    set_rescue_options,
     tf_clean_session,
     tf_disable,
-    tf_login,
-    tf_set_validity_token_cookie,
-    tf_verify_validility_token,
 )
 from .utils import (
     base_render_json,
     check_and_update_authn_fresh,
     config_value as cv,
     do_flash,
+    get_identity_attributes,
     get_message,
     get_post_login_redirect,
     get_post_logout_redirect,
@@ -89,6 +93,7 @@ from .utils import (
     get_post_verify_redirect,
     get_request_attr,
     get_url,
+    hash_password,
     json_error_response,
     login_user,
     logout_user,
@@ -98,6 +103,16 @@ from .utils import (
     url_for_security,
     view_commit,
 )
+from .webauthn import (
+    has_webauthn,
+    webauthn_delete,
+    webauthn_register,
+    webauthn_register_response,
+    webauthn_signin,
+    webauthn_signin_response,
+    webauthn_verify,
+    webauthn_verify_response,
+)
 
 if get_quart_status():  # pragma: no cover
     from quart import make_response, redirect
@@ -151,7 +166,7 @@ def login() -> "ResponseValue":
     form_class = _security.login_form
 
     if request.is_json:
-        # Allow GET so we can return csrf_token for pre-login.
+        # Allow GET, so we can return csrf_token for pre-login.
         if request.content_length:
             form = form_class(MultiDict(request.get_json()), meta=suppress_form_csrf())
         else:
@@ -160,57 +175,66 @@ def login() -> "ResponseValue":
         form = form_class(request.form, meta=suppress_form_csrf())
 
     if form.validate_on_submit():
+        assert form.user is not None
         remember_me = form.remember.data if "remember" in form else None
-        if cv("TWO_FACTOR"):
-            if request.is_json and request.content_length:
-                tf_validity_token = request.get_json().get(  # type: ignore
-                    "tf_validity_token", None
-                )
-            else:
-                tf_validity_token = request.cookies.get("tf_validity", default=None)
-
-            tf_validity_token_is_valid = tf_verify_validility_token(
-                tf_validity_token, form.user.fs_uniquifier
-            )
-
-            if cv("TWO_FACTOR_REQUIRED") or (is_tf_setup(form.user)):
-                if cv("TWO_FACTOR_ALWAYS_VALIDATE") or (not tf_validity_token_is_valid):
-
-                    return tf_login(
-                        form.user, remember=remember_me, primary_authn_via="password"
-                    )
-
-        login_user(form.user, remember=remember_me, authn_via=["password"])
+        response = _security.two_factor_plugins.tf_enter(
+            form.user, remember_me, "password"
+        )
+        if response:
+            return response
+        # two factor not required - login user
         after_this_request(view_commit)
+        login_user(form.user, remember=remember_me, authn_via=["password"])
 
         if _security._want_json(request):
             return base_render_json(form, include_auth_token=True)
         return redirect(get_post_login_redirect())
 
-    if _security._want_json(request):
-        if current_user.is_authenticated:
-            form.user = current_user
-        return base_render_json(form)
+    if (
+        request.method == "POST"
+        and cv("RETURN_GENERIC_RESPONSES")
+        and not form.user_authenticated
+    ):
+        # Validation failed - make sure all error messages are generic
+        fields_to_squash = dict(
+            email=dict(replace_msg="GENERIC_AUTHN_FAILED"),
+            password=dict(replace_msg="GENERIC_AUTHN_FAILED"),
+        )
+        if hasattr(form, "username"):
+            fields_to_squash["username"] = dict(replace_msg="GENERIC_AUTHN_FAILED")
+        form_errors_munge(form, fields_to_squash)
 
     if current_user.is_authenticated:
+        form.user = current_user
+        if _security._want_json(request):
+            return base_render_json(form)
         return redirect(get_url(cv("POST_LOGIN_VIEW")))
-    else:
-        if form.requires_confirmation and cv("REQUIRES_CONFIRMATION_ERROR_VIEW"):
-            do_flash(*get_message("CONFIRMATION_REQUIRED"))
-            return redirect(
-                get_url(
-                    cv("REQUIRES_CONFIRMATION_ERROR_VIEW"),
-                    qparams={"email": form.email.data},
-                )
+
+    if _security._want_json(request):
+        payload = {
+            "identity_attributes": get_identity_attributes(),
+        }
+        return base_render_json(form, additional=payload)
+
+    if form.requires_confirmation and cv("REQUIRES_CONFIRMATION_ERROR_VIEW"):
+        do_flash(*get_message("CONFIRMATION_REQUIRED"))
+        return redirect(
+            get_url(
+                cv("REQUIRES_CONFIRMATION_ERROR_VIEW"),
+                qparams={"email": form.email.data},
             )
-        return _security.render_template(
-            cv("LOGIN_USER_TEMPLATE"), login_user_form=form, **_ctx("login")
         )
+    return _security.render_template(
+        cv("LOGIN_USER_TEMPLATE"),
+        login_user_form=form,
+        identity_attributes=get_identity_attributes(),
+        **_ctx("login"),
+    )
 
 
 @auth_required(lambda: cv("API_ENABLED_METHODS"))
 def verify():
-    """View function which handles a authentication verification request."""
+    """View function which handles a reauthentication request."""
     form_class = _security.verify_form
 
     if request.is_json:
@@ -218,6 +242,7 @@ def verify():
     else:
         form = form_class(meta=suppress_form_csrf())
 
+    form.user = current_user
     if form.validate_on_submit():
         # form may have called verify_and_update_password()
         after_this_request(view_commit)
@@ -226,16 +251,23 @@ def verify():
         session["fs_paa"] = time.time()
 
         if _security._want_json(request):
-            return base_render_json(form)
+            return base_render_json(form, include_auth_token=True)
         do_flash(*get_message("REAUTHENTICATION_SUCCESSFUL"))
         return redirect(get_post_verify_redirect())
 
+    webauthn_available = has_webauthn(current_user, cv("WAN_ALLOW_AS_VERIFY"))
     if _security._want_json(request):
-        assert form.user == current_user
-        return base_render_json(form)
+        payload = {
+            "has_webauthn_verify_credential": webauthn_available,
+        }
+        return base_render_json(form, additional=payload)
 
     return _security.render_template(
-        cv("VERIFY_TEMPLATE"), verify_form=form, **_ctx("verify")
+        cv("VERIFY_TEMPLATE"),
+        verify_form=form,
+        has_webauthn_verify_credential=webauthn_available,
+        wan_verify_form=_security.wan_verify_form(formdata=None),
+        **_ctx("verify"),
     )
 
 
@@ -248,7 +280,7 @@ def logout():
 
     # No body is required - so if a POST and json - return OK
     if request.method == "POST" and _security._want_json(request):
-        return _security._render_json({}, 200, headers=None, user=None)
+        return _security._render_json({}, 200, None, None)
 
     return redirect(get_post_logout_redirect())
 
@@ -273,6 +305,7 @@ def register() -> "ResponseValue":
 
     form = form_class(form_data, meta=suppress_form_csrf())
     if form.validate_on_submit():
+        after_this_request(view_commit)
         did_login = False
         user = register_user(form)
         form.user = user
@@ -282,9 +315,12 @@ def register() -> "ResponseValue":
         # since the beginning. Note that we still enforce 2FA - however for unified
         # signin - we adhere to historic behavior.
         if not _security.confirmable or cv("LOGIN_WITHOUT_CONFIRMATION"):
-            if cv("TWO_FACTOR") and cv("TWO_FACTOR_REQUIRED"):
-                return tf_login(user, primary_authn_via="register")
-            after_this_request(view_commit)
+            response = _security.two_factor_plugins.tf_enter(
+                form.user, False, "register"
+            )
+            if response:
+                return response
+            # two factor not required - login user.
             login_user(user, authn_via=["register"])
             did_login = True
 
@@ -293,6 +329,16 @@ def register() -> "ResponseValue":
 
         # Only include auth token if in fact user is permitted to login
         return base_render_json(form, include_auth_token=did_login)
+
+    # Here on GET or failed validate
+    if request.method == "POST" and cv("RETURN_GENERIC_RESPONSES"):
+        gr = register_existing(form)
+        if gr:
+            if _security._want_json(request):
+                return base_render_json(form)
+
+            return redirect(get_post_register_redirect())
+
     if _security._want_json(request):
         return base_render_json(form)
 
@@ -383,10 +429,20 @@ def send_confirmation():
     if form.validate_on_submit():
         send_confirmation_instructions(form.user)
         if not _security._want_json(request):
-            do_flash(*get_message("CONFIRMATION_REQUEST", email=form.user.email))
+            do_flash(*get_message("CONFIRMATION_REQUEST", email=form.email.data))
+
+    elif request.method == "POST" and cv("RETURN_GENERIC_RESPONSES"):
+        # Here on GET or failed validate
+        rinfo = dict(email=dict())
+        form_errors_munge(form, rinfo)  # by suppressing errors JSON should return 200
+
+        # Make look exactly like successful (e.g. real user) request
+        if not _security._want_json(request):
+            do_flash(*get_message("CONFIRMATION_REQUEST", email=form.email.data))
 
     if _security._want_json(request):
-        return base_render_json(form)
+        # Never include user info since this is an anonymous endpoint.
+        return base_render_json(form, include_user=False)
 
     return _security.render_template(
         cv("SEND_CONFIRMATION_TEMPLATE"),
@@ -447,8 +503,9 @@ def confirm_email(token):
             # and you have the LOGIN_WITH_CONFIRMATION flag since in that case
             # you can be logged in and doing stuff - but another person could
             # get the email.
-            if cv("TWO_FACTOR") and cv("TWO_FACTOR_REQUIRED"):
-                return tf_login(user, primary_authn_via="confirm")
+            response = _security.two_factor_plugins.tf_enter(user, False, "confirm")
+            if response:
+                return response
             login_user(user, authn_via=["confirm"])
 
     m, c = get_message("EMAIL_CONFIRMED")
@@ -484,12 +541,27 @@ def forgot_password():
     if form.validate_on_submit():
         send_reset_password_instructions(form.user)
         if not _security._want_json(request):
-            do_flash(*get_message("PASSWORD_RESET_REQUEST", email=form.user.email))
+            do_flash(*get_message("PASSWORD_RESET_REQUEST", email=form.email.data))
+
+    elif request.method == "POST" and cv("RETURN_GENERIC_RESPONSES"):
+        # Here on GET or failed validate
+        rinfo = dict(email=dict())
+        form_errors_munge(form, rinfo)  # by suppressing errors JSON should return 200
+
+        # Make look exactly like successful (e.g. real user) request
+        hash_password("not-a-password")  # reduce timing between successful and not.
+        if not _security._want_json(request):
+            do_flash(*get_message("PASSWORD_RESET_REQUEST", email=form.email.data))
 
     if _security._want_json(request):
+        # Never include user info since this is an anonymous endpoint.
         return base_render_json(form, include_user=False)
 
-    if form.requires_confirmation and _security.requires_confirmation_error_view:
+    if (
+        form.requires_confirmation
+        and _security.requires_confirmation_error_view
+        and not cv("RETURN_GENERIC_RESPONSES")
+    ):
         do_flash(*get_message("CONFIRMATION_REQUIRED"))
         return redirect(
             get_url(
@@ -592,21 +664,21 @@ def reset_password(token):
 
     if invalid or expired:
         if _security._want_json(request):
-            return _security._render_json(json_error_response(m), 400, None, None)
+            form.form_errors.append(m)
+            return base_render_json(form, include_user=False)
         else:
             return redirect(url_for_security("forgot_password"))
 
     if form.validate_on_submit():
         after_this_request(view_commit)
         update_password(user, form.password.data)
-        if cv("TWO_FACTOR") and (
-            cv("TWO_FACTOR_REQUIRED")
-            or (form.user.tf_totp_secret and form.user.tf_primary_method)
-        ):
-            return tf_login(user, primary_authn_via="reset")
+        response = _security.two_factor_plugins.tf_enter(form.user, False, "reset")
+        if response:
+            return response
+        # two factor not required - just login
         login_user(user, authn_via=["reset"])
         if _security._want_json(request):
-            login_form = _security.login_form()
+            login_form = _security.login_form(formdata=None)
             login_form.user = user
             return base_render_json(login_form, include_auth_token=True)
         else:
@@ -632,11 +704,23 @@ def change_password():
     """View function which handles a change password request."""
 
     form_class = _security.change_password_form
+    form_data = None
+    if request.content_length:
+        form_data = MultiDict(request.get_json()) if request.is_json else request.form
+    form = form_class(formdata=form_data, meta=suppress_form_csrf())
 
-    if request.is_json:
-        form = form_class(MultiDict(request.get_json()), meta=suppress_form_csrf())
-    else:
-        form = form_class(meta=suppress_form_csrf())
+    if not current_user.password:
+        # This is case where user registered w/o a password - since we can't
+        # confirm with existing password - make sure fresh using whatever authentication
+        # method they have set up.
+        if not check_and_update_authn_fresh(
+            cv("FRESHNESS"),
+            cv("FRESHNESS_GRACE_PERIOD"),
+            get_request_attr("fs_authn_via"),
+        ):
+            return _security._reauthn_handler(
+                cv("FRESHNESS"), cv("FRESHNESS_GRACE_PERIOD")
+            )
 
     if form.validate_on_submit():
         after_this_request(view_commit)
@@ -650,13 +734,16 @@ def change_password():
             get_url(_security.post_change_view) or get_url(_security.post_login_view)
         )
 
+    active_password = True if current_user.password else False
     if _security._want_json(request):
         form.user = current_user
-        return base_render_json(form)
+        payload = dict(active_password=active_password)
+        return base_render_json(form, additional=payload)
 
     return _security.render_template(
         cv("CHANGE_PASSWORD_TEMPLATE"),
         change_password_form=form,
+        active_password=active_password,
         **_ctx("change_password"),
     )
 
@@ -699,12 +786,12 @@ def two_factor_setup():
         ] not in ["setup_from_login", "validating_profile"]:
             # illegal call on this endpoint
             tf_clean_session()
-            return _tf_illegal_state(form, _security.login_url)
+            return tf_illegal_state(form, _security.login_url)
 
         user = _datastore.find_user(fs_uniquifier=session["tf_user_id"])
         if not user:
             tf_clean_session()
-            return _tf_illegal_state(form, _security.login_url)
+            return tf_illegal_state(form, _security.login_url)
 
     else:
         # Caller is changing their TFA profile. This requires a 'fresh' authentication
@@ -790,7 +877,7 @@ def two_factor_setup():
                 authr_username=authr_setup_values["username"],
                 authr_issuer=authr_setup_values["issuer"],
             )
-        code_form = _security.two_factor_verify_code_form()
+        code_form = _security.two_factor_verify_code_form(formdata=None)
         if not _security._want_json(request):
             return _security.render_template(
                 cv("TWO_FACTOR_SETUP_TEMPLATE"),
@@ -820,7 +907,7 @@ def two_factor_setup():
         }
         return base_render_json(form, include_user=False, additional=json_response)
 
-    code_form = _security.two_factor_verify_code_form()
+    code_form = _security.two_factor_verify_code_form(formdata=None)
     return _security.render_template(
         cv("TWO_FACTOR_SETUP_TEMPLATE"),
         two_factor_setup_form=form,
@@ -845,11 +932,10 @@ def two_factor_token_validation():
     """
 
     form_class = _security.two_factor_verify_code_form
-
-    if request.is_json:
-        form = form_class(MultiDict(request.get_json()), meta=suppress_form_csrf())
-    else:
-        form = form_class(meta=suppress_form_csrf())
+    form_data = None
+    if request.content_length:
+        form_data = MultiDict(request.get_json()) if request.is_json else request.form
+    form = form_class(formdata=form_data, meta=suppress_form_csrf())
 
     changing = current_user.is_authenticated
     if not changing:
@@ -866,18 +952,20 @@ def two_factor_token_validation():
         ):
             # illegal call on this endpoint
             tf_clean_session()
-            return _tf_illegal_state(form, _security.login_url)
+            return tf_illegal_state(form, _security.login_url)
 
         user = _datastore.find_user(fs_uniquifier=session["tf_user_id"])
         form.user = user
         if not user:
             tf_clean_session()
-            return _tf_illegal_state(form, _security.login_url)
+            return tf_illegal_state(form, _security.login_url)
 
         if session["tf_state"] == "ready":
+            # normal login case
             pm = user.tf_primary_method
             totp_secret = user.tf_totp_secret
         else:
+            # initial setup case
             pm = session["tf_primary_method"]
             totp_secret = session["tf_totp_secret"]
     else:
@@ -889,7 +977,7 @@ def two_factor_token_validation():
             tf_clean_session()
             # logout since this seems like attack-ish/logic error
             logout_user()
-            return _tf_illegal_state(form, _security.login_url)
+            return tf_illegal_state(form, _security.login_url)
         pm = session["tf_primary_method"]
         totp_secret = session["tf_totp_secret"]
         form.user = current_user
@@ -898,55 +986,54 @@ def two_factor_token_validation():
     form.tf_totp_secret = totp_secret
     if form.validate_on_submit():
         # Success - log in user and clear all session variables
-        remember = session.pop("tf_remember_login", None)
-        completion_message = complete_two_factor_process(
-            form.user, pm, totp_secret, changing, remember
+        completion_message, token = complete_two_factor_process(
+            form.user, pm, totp_secret, changing
         )
 
         after_this_request(view_commit)
 
         if not _security._want_json(request):
-            after_this_request(
-                partial(
-                    tf_set_validity_token_cookie,
-                    fs_uniquifier=form.user.fs_uniquifier,
-                    remember=remember,
-                )
-            )
+            if token:
+                after_this_request(partial(tf_set_validity_token_cookie, token=token))
             do_flash(*get_message(completion_message))
-
             return redirect(get_post_login_redirect())
 
-        if (not cv("TWO_FACTOR_ALWAYS_VALIDATE") and remember) and _security._want_json(
-            request
-        ):
-            token = generate_tf_validity_token(form.user.fs_uniquifier)
-            json_response = {"tf_validity_token": token}
-            return base_render_json(form, additional=json_response)
+        else:
+            json_payload = {}
+            if token:
+                json_payload["tf_validity_token"] = token
+            return base_render_json(form, additional=json_payload)
+
     # GET or not successful POST
-    if _security._want_json(request):
-        return base_render_json(form)
 
     # if we were trying to validate a new method
     if changing:
-        setup_form = _security.two_factor_setup_form()
+        if _security._want_json(request):
+            return base_render_json(form)
+        setup_form = _security.two_factor_setup_form(formdata=None)
 
         return _security.render_template(
             cv("TWO_FACTOR_SETUP_TEMPLATE"),
             two_factor_setup_form=setup_form,
             two_factor_verify_code_form=form,
+            chosen_method=pm,
             choices=cv("TWO_FACTOR_ENABLED_METHODS"),
             **_ctx("tf_setup"),
         )
 
     # if we were trying to validate an existing method
     else:
-        rescue_form = _security.two_factor_rescue_form()
-
+        rescue_form = _security.two_factor_rescue_form(formdata=None)
+        recovery_options = set_rescue_options(rescue_form, form.user)
+        if _security._want_json(request):
+            return base_render_json(
+                form, additional=dict(recovery_options=recovery_options)
+            )
         return _security.render_template(
             cv("TWO_FACTOR_VERIFY_CODE_TEMPLATE"),
             two_factor_rescue_form=rescue_form,
             two_factor_verify_code_form=form,
+            chosen_method=pm,
             problem=None,
             **_ctx("tf_token_validation"),
         )
@@ -964,32 +1051,21 @@ def two_factor_rescue():
     """
 
     form_class = _security.two_factor_rescue_form
+    form_data = None
+    if request.content_length:
+        form_data = MultiDict(request.get_json()) if request.is_json else request.form
+    form = form_class(formdata=form_data, meta=suppress_form_csrf())
 
-    if request.is_json:
-        form = form_class(MultiDict(request.get_json()), meta=suppress_form_csrf())
-    else:
-        form = form_class(meta=suppress_form_csrf())
-
-    if (
-        not all(k in session for k in ["tf_user_id", "tf_state"])
-        or session["tf_state"] != "ready"
-    ):
-        tf_clean_session()
-        return _tf_illegal_state(form, _security.login_url)
-
-    user = _datastore.find_user(fs_uniquifier=session["tf_user_id"])
-    form.user = user
-    if not user:
-        tf_clean_session()
-        return _tf_illegal_state(form, _security.login_url)
+    form.user = tf_check_state(["ready"])
+    if not form.user:
+        return tf_illegal_state(form, _security.login_url)
 
+    recovery_options = set_rescue_options(form, form.user)
     rproblem = ""
     if form.validate_on_submit():
-        problem = form.data["help_setup"]
-        rproblem = problem
-        # if the problem is that user can't access his device, w
-        # e send him code through mail
-        if problem == "lost_device":
+        raction = form.help_setup.data
+        rproblem = raction
+        if raction == "email":
             msg = form.user.tf_send_security_token(
                 method="email",
                 totp_secret=form.user.tf_totp_secret,
@@ -1002,41 +1078,125 @@ def two_factor_rescue():
                     return base_render_json(
                         form, include_user=False, error_status_code=500
                     )
+            # drop through to GET path
+        elif raction == "recovery_code":
+            return redirect(url_for_security("mf_recovery"))
+
         # send app provider a mail message regarding trouble
-        elif problem == "no_mail_access":
+        elif raction == "help":
             send_mail(
                 cv("EMAIL_SUBJECT_TWO_FACTOR_RESCUE"),
                 cv("TWO_FACTOR_RESCUE_MAIL"),
                 "two_factor_rescue",
                 user=form.user,
             )
+            # drop through to GET path
         else:
             return "", 404
 
     if _security._want_json(request):
-        return base_render_json(form, include_user=False)
+        return base_render_json(
+            form, include_user=False, additional=dict(recovery_options=recovery_options)
+        )
 
-    code_form = _security.two_factor_verify_code_form()
+    code_form = _security.two_factor_verify_code_form(formdata=None)
     return _security.render_template(
         cv("TWO_FACTOR_VERIFY_CODE_TEMPLATE"),
         two_factor_verify_code_form=code_form,
         two_factor_rescue_form=form,
+        chosen_method=form.user.tf_primary_method,
         rescue_mail=cv("TWO_FACTOR_RESCUE_MAIL"),
         problem=rproblem,
         **_ctx("tf_token_validation"),
     )
 
 
-def _tf_illegal_state(form, redirect_to):
-    m, c = get_message("TWO_FACTOR_PERMISSION_DENIED")
-    if not _security._want_json(request):
-        do_flash(m, c)
-        return redirect(get_url(redirect_to))
-    else:
-        return _security._render_json(json_error_response(m), 400, None, None)
+@auth_required(
+    lambda: cv("API_ENABLED_METHODS"),
+    within=lambda: cv("FRESHNESS"),
+    grace=lambda: cv("FRESHNESS_GRACE_PERIOD"),
+)
+def mf_recovery_codes() -> "ResponseValue":
+    """
+    Create and download multi-factor recovery codes.
+    For forms we want the user to explicitly request to see the codes - so
+    the form has a show_codes submit button.
+    """
+    form_class = _security.mf_recovery_codes_form
+    form_data = None
+    if request.content_length:
+        form_data = MultiDict(request.get_json()) if request.is_json else request.form
+    form = form_class(formdata=form_data, meta=suppress_form_csrf())
+
+    if form.validate_on_submit():
+        # generate new codes
+        codes = create_recovery_codes(current_user)
+        after_this_request(view_commit)
+        if _security._want_json(request):
+            payload = dict(recovery_codes=codes)
+            return base_render_json(form, include_user=False, additional=payload)
+        return _security.render_template(
+            cv("MULTI_FACTOR_RECOVERY_CODES_TEMPLATE"),
+            mf_recovery_codes_form=form,
+            recovery_codes=codes,
+            **_ctx("mf_recovery_codes"),
+        )
+
+    codes = _datastore.mf_get_recovery_codes(current_user)
+    if _security._want_json(request):
+        return base_render_json(
+            form, include_user=False, additional=dict(recovery_codes=codes)
+        )
+    return _security.render_template(
+        cv("MULTI_FACTOR_RECOVERY_CODES_TEMPLATE"),
+        mf_recovery_codes_form=form,
+        recovery_codes=codes if request.args.get("show_codes", None) else [],
+        **_ctx("mf_recovery_codes"),
+    )
 
 
-def create_blueprint(app, state, import_name, json_encoder=None):
+@anonymous_user_required
+@unauth_csrf(fall_through=True)
+def mf_recovery():
+    """View for entering a recovery code.
+
+    User must have already provided valid username/password.
+    User must have already established 2FA
+
+    """
+    form_class = _security.mf_recovery_form
+    form_data = None
+    if request.content_length:
+        form_data = MultiDict(request.get_json()) if request.is_json else request.form
+    form = form_class(formdata=form_data, meta=suppress_form_csrf())
+
+    form.user = tf_check_state(["ready"])
+    if not form.user:
+        return tf_illegal_state(form, _security.login_url)
+
+    if form.validate_on_submit():
+        # Valid code - we want these to be one time - so remove it from list
+        _datastore.mf_delete_recovery_code(form.user, form.code.data)
+        after_this_request(view_commit)
+
+        # In the recovery case - don't set/offer validity token.
+        _security.two_factor_plugins.tf_complete(form.user, True)
+
+        if not _security._want_json(request):
+            return redirect(get_post_login_redirect())
+        else:
+            return base_render_json(form)
+
+    if _security._want_json(request):
+        return base_render_json(form, include_user=False)
+    return _security.render_template(
+        cv("MULTI_FACTOR_RECOVERY_TEMPLATE"),
+        mf_recovery_form=form,
+        **_ctx("mf_recovery"),
+    )
+
+
+def create_blueprint(app, state, import_name):
     """Creates the security extension blueprint"""
 
     bp = Blueprint(
@@ -1046,8 +1206,6 @@ def create_blueprint(app, state, import_name, json_encoder=None):
         subdomain=state.subdomain,
         template_folder="templates",
     )
-    if json_encoder:
-        bp.json_encoder = json_encoder
 
     if state.logout_methods is not None:
         bp.route(state.logout_url, methods=state.logout_methods, endpoint="logout")(
@@ -1075,7 +1233,7 @@ def create_blueprint(app, state, import_name, json_encoder=None):
         )
         bp.route(
             state.us_signin_send_code_url,
-            methods=["GET", "POST"],
+            methods=["POST"],
             endpoint="us_signin_send_code",
         )(us_signin_send_code)
         bp.route(state.us_setup_url, methods=["GET", "POST"], endpoint="us_setup")(
@@ -1094,7 +1252,7 @@ def create_blueprint(app, state, import_name, json_encoder=None):
             )(us_verify)
             bp.route(
                 state.us_verify_send_code_url,
-                methods=["GET", "POST"],
+                methods=["POST"],
                 endpoint="us_verify_send_code",
             )(us_verify_send_code)
 
@@ -1150,4 +1308,52 @@ def create_blueprint(app, state, import_name, json_encoder=None):
             endpoint="confirm_email",
         )(confirm_email)
 
+    if cv("MULTI_FACTOR_RECOVERY_CODES", app) and state.support_mfa:
+        bp.route(
+            state.multi_factor_recovery_codes_url,
+            methods=["GET", "POST"],
+            endpoint="mf_recovery_codes",
+        )(mf_recovery_codes)
+        bp.route(
+            state.multi_factor_recovery_url,
+            methods=["GET", "POST"],
+            endpoint="mf_recovery",
+        )(mf_recovery)
+
+    if state.webauthn:
+        bp.route(
+            state.wan_register_url,
+            methods=["GET", "POST"],
+            endpoint="wan_register",
+        )(webauthn_register)
+        bp.route(
+            state.wan_register_url
+            + slash_url_suffix(state.wan_register_url, "<token>"),
+            methods=["POST"],
+            endpoint="wan_register_response",
+        )(webauthn_register_response)
+        bp.route(state.wan_signin_url, methods=["GET", "POST"], endpoint="wan_signin")(
+            webauthn_signin
+        )
+        bp.route(
+            state.wan_signin_url + slash_url_suffix(state.wan_signin_url, "<token>"),
+            methods=["POST"],
+            endpoint="wan_signin_response",
+        )(webauthn_signin_response)
+        bp.route(state.wan_delete_url, methods=["GET", "POST"], endpoint="wan_delete")(
+            webauthn_delete
+        )
+        if cv("FRESHNESS", app=app).total_seconds() >= 0 and cv(
+            "WAN_ALLOW_AS_VERIFY", app=app
+        ):
+            bp.route(
+                state.wan_verify_url, methods=["GET", "POST"], endpoint="wan_verify"
+            )(webauthn_verify)
+            bp.route(
+                state.wan_verify_url
+                + slash_url_suffix(state.wan_verify_url, "<token>"),
+                methods=["POST"],
+                endpoint="wan_verify_response",
+            )(webauthn_verify_response)
+
     return bp
