@@ -4,7 +4,7 @@
 
     Test common functionality
 
-    :copyright: (c) 2019 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
@@ -113,9 +113,11 @@ def test_authenticate_case_insensitive_email(app, client):
 
 def test_authenticate_with_invalid_input(client, get_message):
     response = client.post(
-        "/login", data="{}", headers={"Content-Type": "application/json"}
+        "/login",
+        json=dict(password="password"),
+        headers={"Content-Type": "application/json"},
     )
-    assert get_message("EMAIL_NOT_PROVIDED") in response.data
+    assert get_message("USER_DOES_NOT_EXIST") in response.data
 
 
 @pytest.mark.settings(post_login_view="/post_login")
@@ -155,16 +157,80 @@ def test_login_form(client):
 
 @pytest.mark.settings(username_enable=True)
 def test_login_form_username(client):
-    # If USERNAME_ENABLE is set then login form should have a StringField, not Email
-    response = client.post("/login", data={"email": "matt@lp.com"})
-    assert b"matt@lp.com" in response.data
-    # Should be a string field - not html5 email input type
-    assert not re.search(b'<input[^>]*type="email"[^>]*>', response.data)
+    # If USERNAME_ENABLE is set then login form should have a both an Email and
+    # StringField
+    response = client.get("/login")
+    # Should be both email with Email type and username with autocomplete
+    assert re.search(b'<input[^>]*type="email"[^>]*>', response.data)
+    assert re.search(b'<input[^>]*autocomplete="username"[^>]*>', response.data)
+    assert re.search(b'<input[^>]*autocomplete="current-password"[^>]*>', response.data)
+
+
+@pytest.mark.confirmable()
+@pytest.mark.settings(
+    return_generic_responses=True, requires_confirmation_error_view="/confirm"
+)
+def test_generic_response(app, client, get_message):
+    response = client.post(
+        "/login", data=dict(email="mattwho@lp.com", password="forgot")
+    )
+    assert get_message("GENERIC_AUTHN_FAILED") in response.data
+    response = client.post("/login", data=dict(email="matt@lp.com", password="forgot"))
+    assert get_message("GENERIC_AUTHN_FAILED") in response.data
+
+    response = client.post(
+        "/login", json=dict(email="mattwho@lp.com", password="forgot")
+    )
+    # make sure just 'null' key in errors.
+    assert list(response.json["response"]["field_errors"].keys()) == ["null"]
+    assert len(response.json["response"]["field_errors"]["null"]) == 1
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
+        "GENERIC_AUTHN_FAILED"
+    )
+    response = client.post("/login", json=dict(email="matt@lp.com", password="forgot"))
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
+        "GENERIC_AUTHN_FAILED"
+    )
+
+    # make sure don't get confirmation required
+    response = client.post(
+        "/login",
+        data=dict(email="mattwho@lp.com", password="password"),
+        follow_redirects=False,
+    )
+    assert response.status_code == 200
+
+
+@pytest.mark.registerable()
+@pytest.mark.settings(username_enable=True, return_generic_responses=True)
+def test_generic_response_username(app, client, get_message):
+    data = dict(
+        email="dude@lp.com",
+        username="dude",
+        password="awesome sunset",
+    )
+    response = client.post("/register", json=data)
+    assert response.headers["Content-Type"] == "application/json"
+    assert response.status_code == 200
+    logout(client)
+
+    response = client.post(
+        "/login", data=dict(username="dude2", password="awesome sunset")
+    )
+    assert get_message("GENERIC_AUTHN_FAILED") in response.data
+
+    response = client.post("/login", json=dict(username="dude2", password="forgot"))
+    # make sure just 'null' key in errors.
+    assert list(response.json["response"]["field_errors"].keys()) == ["null"]
+    assert len(response.json["response"]["field_errors"]["null"]) == 1
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
+        "GENERIC_AUTHN_FAILED"
+    )
 
 
 def test_unprovided_username(client, get_message):
     response = authenticate(client, "")
-    assert get_message("EMAIL_NOT_PROVIDED") in response.data
+    assert get_message("USER_DOES_NOT_EXIST") in response.data
 
 
 def test_unprovided_password(client, get_message):
@@ -263,7 +329,9 @@ def test_inactive_forbids_basic(app, client, get_message):
 
 def test_unset_password(client, get_message):
     response = authenticate(client, "jess@lp.com", "password")
-    assert get_message("PASSWORD_NOT_SET") in response.data
+    assert get_message("INVALID_PASSWORD") in response.data
+    response = authenticate(client, "jess@lp.com", "")
+    assert get_message("PASSWORD_NOT_PROVIDED") in response.data
 
 
 def test_logout(client):
@@ -510,7 +578,7 @@ def test_http_auth_no_authentication(client, get_message):
 def test_http_auth_no_authentication_json(client, get_message):
     response = client.get("/http", headers={"accept": "application/json"})
     assert response.status_code == 401
-    assert response.json["response"]["error"].encode("utf-8") == get_message(
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
         "UNAUTHENTICATED"
     )
     assert response.headers["Content-Type"] == "application/json"
@@ -542,7 +610,7 @@ def test_invalid_http_auth_invalid_username_json(client, get_message):
         },
     )
     assert response.status_code == 401
-    assert response.json["response"]["error"].encode("utf-8") == get_message(
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
         "UNAUTHENTICATED"
     )
     assert response.headers["Content-Type"] == "application/json"
@@ -734,7 +802,7 @@ def test_anon_required_json(client, get_message):
     authenticate(client, follow_redirects=False)
     response = client.get("/register", headers={"Accept": "application/json"})
     assert response.status_code == 400
-    assert response.json["response"]["error"].encode("utf-8") == get_message(
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
         "ANONYMOUS_USER_REQUIRED"
     )
 
@@ -950,6 +1018,7 @@ def test_auth_token_decorator(in_app_context):
     client_nc = app.test_client(use_cookies=False)
 
     with app.test_request_context("/"):
+
         user = app.security.datastore.find_user(email="matt@lp.com")
         token = get_auth_token_version_3x(app, user)
 
