@@ -69,7 +69,7 @@ def test_confirmable_flag(app, clients, get_message):
     )
     assert response.status_code == 200
     assert response.headers["Content-Type"] == "application/json"
-    assert "user" in response.json["response"]
+    assert "user" not in response.json["response"]
     assert len(recorded_instructions_sent) == 1
 
     # Test ask for instructions with invalid email
@@ -94,8 +94,9 @@ def test_confirmable_flag(app, clients, get_message):
 
     # Test already confirmed and expired token
     app.config["SECURITY_CONFIRM_EMAIL_WITHIN"] = "-1 days"
-    with app.app_context():
-        user = registrations[0]["user"]
+    with app.test_request_context("/"):
+        email = registrations[0]["email"]
+        user = app.security.datastore.find_user(email=email)
         expired_token = generate_confirmation_token(user)
     response = clients.get("/confirm/" + expired_token, follow_redirects=True)
     assert get_message("ALREADY_CONFIRMED") in response.data
@@ -141,13 +142,13 @@ def test_confirmation_template(app, client, get_message):
         recorded_tokens_sent.append(kwargs["confirmation_token"])
 
     with capture_registrations() as registrations:
-        with app.mail.record_messages() as outbox:
-            data = dict(email="mary@lp.com", password="awesome sunset", next="")
-            # Register - this will use the welcome template
-            client.post("/register", data=data, follow_redirects=True)
-            # Explicitly ask for confirmation -
-            # this will use the confirmation_instructions template
-            client.post("/confirm", data=dict(email="mary@lp.com"))
+        data = dict(email="mary@lp.com", password="awesome sunset", next="")
+        # Register - this will use the welcome template
+        client.post("/register", data=data, follow_redirects=True)
+        # Explicitly ask for confirmation -
+        # this will use the confirmation_instructions template
+        client.post("/confirm", data=dict(email="mary@lp.com"))
+        outbox = app.mail.outbox
         assert len(outbox) == 2
         # check registration email
         matcher = re.findall(r"\w+:.*", outbox[0].body, re.IGNORECASE)
@@ -190,13 +191,13 @@ def test_expired_confirmation_token(client, get_message):
         data = dict(email="mary@lp.com", password="awesome sunset", next="")
         client.post("/register", data=data, follow_redirects=True)
 
-    user = registrations[0]["user"]
+    email = registrations[0]["email"]
     token = registrations[0]["confirm_token"]
 
     time.sleep(1)
 
     response = client.get("/confirm/" + token, follow_redirects=True)
-    msg = get_message("CONFIRMATION_EXPIRED", within="1 milliseconds", email=user.email)
+    msg = get_message("CONFIRMATION_EXPIRED", within="1 milliseconds", email=email)
     assert msg in response.data
 
 
@@ -286,10 +287,7 @@ def test_confirmation_different_user_when_logged_in_no_auto(client, get_message)
     response = client.get("/confirm/" + token2, follow_redirects=True)
     assert get_message("EMAIL_CONFIRMED") in response.data
     # should get a login view
-    assert (
-        b'<input id="password" name="password" required type="password" value="">'
-        in response.data
-    )
+    assert b'<a href="/login">Login</a>' in response.data
 
 
 @pytest.mark.registerable()
@@ -481,7 +479,7 @@ def test_two_factor_json(app, client, get_message):
     # make sure not logged in
     response = client.get("/profile", headers={"accept": "application/json"})
     assert response.status_code == 401
-    assert response.json["response"]["error"].encode("utf-8") == get_message(
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
         "UNAUTHENTICATED"
     )
 
@@ -496,24 +494,10 @@ def test_two_factor_json(app, client, get_message):
 @pytest.mark.registerable()
 @pytest.mark.settings(
     user_identity_attributes=[{"username": {"mapper": lambda x: x}}],
+    username_enable=True,
 )
-def test_email_not_identity(app, sqlalchemy_datastore, get_message):
+def test_email_not_identity(app, client, get_message):
     # Test that can register/confirm with email even if it isn't an IDENTITY_ATTRIBUTE
-    from flask_security import ConfirmRegisterForm, Security, unique_identity_attribute
-    from wtforms import StringField, validators
-
-    class MyRegisterForm(ConfirmRegisterForm):
-        username = StringField(
-            "Username",
-            validators=[validators.data_required(), unique_identity_attribute],
-        )
-
-    app.config["SECURITY_CONFIRM_REGISTER_FORM"] = MyRegisterForm
-    security = Security(datastore=sqlalchemy_datastore)
-    security.init_app(app)
-
-    client = app.test_client()
-
     with capture_registrations() as registrations:
         data = dict(email="mary2@lp.com", username="mary", password="awesome sunset")
         response = client.post("/register", data=data, follow_redirects=True)
@@ -532,13 +516,58 @@ def test_email_not_identity(app, sqlalchemy_datastore, get_message):
         "/register", data=data, headers={"Accept": "application/json"}
     )
     assert response.status_code == 400
-    assert "is already associated" in response.json["response"]["errors"]["username"][0]
+    assert "is already associated" in response.json["response"]["errors"][0]
 
-    # log in with username - this uses the age-old hack that although the form's
-    # input label says "email" - it in fact will accept any identity attribute.
+    # verify that email field not present
+    response = client.get("/login")  # this one has a flash containing 'email confirmed'
+    response = client.get("/login")
+    assert b"email" not in response.data
+    assert b"username" in response.data
+
+    response = client.get("/login", headers={"Content-Type": "application/json"})
+    assert response.json["response"]["identity_attributes"] == ["username"]
+
+    # log in with username
     response = client.post(
         "/login",
-        data=dict(email="mary", password="awesome sunset"),
+        data=dict(username="mary", password="awesome sunset"),
         follow_redirects=True,
     )
     assert b"<p>Welcome mary</p>" in response.data
+
+
+@pytest.mark.settings(return_generic_responses=True)
+def test_generic_response(app, client, get_message):
+    # Confirm matt - then an unknown email - both should get the same answer and
+    # JSON should return 200
+    recorded_instructions_sent = []
+
+    @confirm_instructions_sent.connect_via(app)
+    def on_instructions_sent(app, **kwargs):
+        recorded_instructions_sent.append(kwargs["token"])
+
+    response = client.post("/confirm", data=dict(email="matt@lp.com"))
+    assert len(recorded_instructions_sent) == 1
+    assert get_message("CONFIRMATION_REQUEST", email="matt@lp.com") in response.data
+    response = client.post("/confirm", json=dict(email="matt@lp.com"))
+    assert len(recorded_instructions_sent) == 2
+    assert response.status_code == 200
+
+    # actually confirm matt
+    token = recorded_instructions_sent[0]
+    response = client.get("/confirm/" + token, follow_redirects=True)
+    assert get_message("EMAIL_CONFIRMED") in response.data
+
+    # Try to confirm an unknown email - should get SAME message as real email.
+    response = client.post("/confirm", data=dict(email="mattwho@lp.com"))
+    assert len(recorded_instructions_sent) == 2
+    assert get_message("CONFIRMATION_REQUEST", email="mattwho@lp.com") in response.data
+    response = client.post("/confirm", json=dict(email="mattwho@lp.com"))
+    assert len(recorded_instructions_sent) == 2
+    assert response.status_code == 200
+    assert not any(e in response.json["response"].keys() for e in ["error", "errors"])
+
+    # Try to confirm matt again - should ALSO get same response.
+    response = client.post("/confirm", json=dict(email="matt@lp.com"))
+    assert len(recorded_instructions_sent) == 2
+    assert response.status_code == 200
