@@ -5,12 +5,12 @@
     Datastore tests
 
     :copyright: (c) 2012 by Matt Wright.
-    :copyright: (c) 2019 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
 import datetime
-from pytest import raises, skip
+from pytest import raises, skip, importorskip
 from tests.test_utils import init_app_with_options, get_num_queries, is_sqlalchemy
 
 from flask_security import RoleMixin, Security, UserMixin, LoginForm, RegisterForm
@@ -46,7 +46,7 @@ def test_unimplemented_datastore_methods():
 def test_unimplemented_user_datastore_methods():
     datastore = UserDatastore(None, None)
     with raises(NotImplementedError):
-        datastore.find_user(None)
+        datastore.find_user()
     with raises(NotImplementedError):
         datastore.find_role(None)
 
@@ -233,24 +233,8 @@ def test_create_user_with_roles_and_permissions(app, datastore):
         assert user.has_permission("write") is False
 
 
-def test_permissions_strings(app, datastore):
-    # Make sure spaces are squashed.
-    ds = datastore
-    if not hasattr(ds.role_model, "permissions"):
-        return
-    init_app_with_options(app, ds)
-
-    with app.app_context():
-        perms = "read, write "
-        ds.create_role(name="test1", permissions=perms)
-        ds.commit()
-
-        t1 = ds.find_role("test1")
-        assert {"read", "write"} == t1.get_permissions()
-
-
-def test_permissions_iter(app, datastore):
-    # Test permissions as an interable
+def test_permissions_types(app, datastore):
+    # Test permissions as a list, set, tuple, comma separated list
     ds = datastore
     if not hasattr(ds.role_model, "permissions"):
         return
@@ -264,6 +248,27 @@ def test_permissions_iter(app, datastore):
         t1 = ds.find_role("test1")
         assert {"read", "write"} == t1.get_permissions()
 
+        perms = {"read", "write"}
+        ds.create_role(name="test2", permissions=perms)
+        ds.commit()
+
+        t2 = ds.find_role("test2")
+        assert {"read", "write"} == t2.get_permissions()
+
+        perms = "read, write"
+        ds.create_role(name="test3", permissions=perms)
+        ds.commit()
+
+        t3 = ds.find_role("test3")
+        assert {"read", "write"} == t3.get_permissions()
+
+        perms = ("read", "write")
+        ds.create_role(name="test4", permissions=perms)
+        ds.commit()
+
+        t4 = ds.find_role("test4")
+        assert {"read", "write"} == t4.get_permissions()
+
 
 def test_modify_permissions(app, datastore):
     ds = datastore
@@ -313,7 +318,6 @@ def test_modify_permissions_multi(app, datastore):
     if not hasattr(ds.role_model, "permissions"):
         return
 
-    # N.B. right now only sqlalchemy has the extended RoleModel.
     init_app_with_options(app, ds)
 
     with app.app_context():
@@ -337,11 +341,12 @@ def test_modify_permissions_multi(app, datastore):
 
         # send in a set
         perms = {"read", "write"}
-        ds.create_role(name="test2", permissions=",".join(perms))
+        ds.create_role(name="test2", permissions=perms)
         ds.commit()
 
+        # add permissions using comma separate string
         t2 = ds.find_role("test2")
-        ds.add_permissions_to_role(t2, {"execute", "whatever"})
+        ds.add_permissions_to_role(t2, "execute, whatever")
         ds.commit()
 
         t2 = ds.find_role("test2")
@@ -351,29 +356,15 @@ def test_modify_permissions_multi(app, datastore):
         ds.commit()
         assert {"write", "execute"} == t2.get_permissions()
 
-
-def test_modify_permissions_unsupported(app, datastore):
-    from tests.conftest import PonyUserDatastore
-
-    ds = datastore
-    if hasattr(datastore.role_model, "permissions"):
-        # already tested this
-        return
-    if isinstance(datastore, PonyUserDatastore):
-        # sigh - Pony doesn't use RoleMixin.
-        return
-
-    init_app_with_options(app, ds)
-
-    with app.app_context():
-        ds.create_role(name="test3")
+        ds.remove_permissions_from_role(t2, "write, execute")
         ds.commit()
-        t3 = ds.find_role("test3")
+        assert t2.get_permissions() == set()
 
-        with raises(NotImplementedError):
-            t3.add_permissions("whatever")
-        with raises(NotImplementedError):
-            t3.remove_permissions("whatever")
+        # add permissions using a tuple
+        t2 = ds.find_role("test2")
+        ds.add_permissions_to_role(t2, ("execute2", "whatever2"))
+        ds.commit()
+        assert {"whatever2", "execute2"} == t2.get_permissions()
 
 
 def test_uuid(app, request, tmpdir, realdburl):
@@ -436,9 +427,8 @@ def test_uuid(app, request, tmpdir, realdburl):
         db.create_all()
 
     def tear_down():
-        with app.app_context():
-            db.drop_all()
-            _teardown_realdb(db_info)
+        db.drop_all()
+        _teardown_realdb(db_info)
 
     request.addfinalizer(tear_down)
 
@@ -448,3 +438,121 @@ def test_uuid(app, request, tmpdir, realdburl):
     with app.app_context():
         user = ds.find_user(email="matt@lp.com")
         assert not user
+
+
+def test_webauthn(app, datastore):
+    importorskip("webauthn")
+    if not datastore.webauthn_model:
+        skip("No WebAuthn model defined")
+    init_app_with_options(app, datastore)
+
+    with app.app_context():
+        user = datastore.find_user(email="matt@lp.com")
+        datastore.create_webauthn(
+            user,
+            name="cred1",
+            credential_id=b"1",
+            public_key=b"key",
+            sign_count=0,
+            transports=None,
+            extensions=None,
+            usage="first",
+            device_type="single_device",
+            backup_state=False,
+        )
+        datastore.commit()
+        cred = datastore.find_webauthn(b"1")
+        assert cred.name == "cred1"
+
+        user = datastore.find_user(email="matt@lp.com")
+        assert len(user.webauthn) == 1
+        assert user.webauthn[0].name == "cred1"
+
+        datastore.delete_webauthn(user.webauthn[0])
+        datastore.commit()
+        user = datastore.find_user(email="matt@lp.com")
+        assert len(user.webauthn) == 0
+
+
+def test_webauthn_cascade(app, datastore):
+    importorskip("webauthn")
+    if not datastore.webauthn_model:
+        skip("No WebAuthn model defined")
+    init_app_with_options(app, datastore)
+
+    with app.app_context():
+        user = datastore.find_user(email="matt@lp.com")
+        datastore.create_webauthn(
+            user,
+            name="cred1",
+            credential_id=b"1",
+            public_key=b"key",
+            sign_count=0,
+            transports=None,
+            extensions=None,
+            usage="first",
+            device_type="single_device",
+            backup_state=False,
+        )
+        datastore.create_webauthn(
+            user,
+            name="cred2",
+            credential_id=b"2",
+            public_key=b"key",
+            sign_count=0,
+            transports=None,
+            extensions=None,
+            usage="secondary",
+            device_type="single_device",
+            backup_state=False,
+        )
+        datastore.commit()
+
+        user = datastore.find_user(email="matt@lp.com")
+        assert len(user.webauthn) == 2
+        names = [cred.name for cred in user.webauthn]
+        assert set(names) == {"cred1", "cred2"}
+        assert datastore.find_webauthn(b"1")
+
+        # delete user
+        datastore.delete_user(user)
+        datastore.commit()
+        user = datastore.find_user(email="matt@lp.com")
+        assert not user
+
+        cred = datastore.find_webauthn(b"1")
+        assert not cred
+        cred = datastore.find_webauthn(b"2")
+        assert not cred
+
+
+def test_mf_recovery_codes(app, datastore):
+    from tests.conftest import PonyUserDatastore
+
+    if isinstance(datastore, PonyUserDatastore):
+        skip("Pony not supported")
+    init_app_with_options(app, datastore)
+
+    with app.test_request_context("/"):
+        user = datastore.find_user(email="matt@lp.com")
+        assert hasattr(user, "mf_recovery_codes")
+
+        assert not datastore.mf_delete_recovery_code(user, "nothing yet")
+
+        datastore.mf_set_recovery_codes(user, ["r1", "r2", "r3"])
+        datastore.commit()
+
+        user = datastore.find_user(email="matt@lp.com")
+        codes = datastore.mf_get_recovery_codes(user)
+        assert codes == ["r1", "r2", "r3"]
+
+        rv = datastore.mf_delete_recovery_code(user, "r2")
+        assert rv
+        datastore.commit()
+
+        rv = datastore.mf_delete_recovery_code(user, "huh?")
+        assert not rv
+
+        user = datastore.find_user(email="matt@lp.com")
+        codes = datastore.mf_get_recovery_codes(user)
+        assert codes == ["r1", "r3"]
