@@ -27,6 +27,7 @@ from tests.test_utils import (
     capture_flashes,
     get_session,
     logout,
+    reset_fresh,
 )
 
 pytestmark = pytest.mark.two_factor()
@@ -78,6 +79,7 @@ def tf_in_session(session):
             "tf_user_id",
             "tf_remember_login",
             "tf_totp_secret",
+            "tf_select",
         ]
     )
 
@@ -261,6 +263,7 @@ def test_tf_reset_invalidates_cookie(app, client):
 
 @pytest.mark.settings(two_factor_required=True)
 def test_two_factor_two_factor_setup_anonymous(app, client, get_message):
+
     # trying to pick method without doing earlier stage
     data = dict(setup="email")
 
@@ -274,7 +277,7 @@ def test_two_factor_two_factor_setup_anonymous(app, client, get_message):
 
 
 @pytest.mark.settings(two_factor_required=True)
-def test_two_factor_flag(app, client):
+def test_two_factor_flag(app, client, get_message):
     # trying to verify code without going through two-factor
     # first login function
     wrong_code = b"000000"
@@ -331,7 +334,8 @@ def test_two_factor_flag(app, client):
     )
     assert response.status_code == 400
     assert (
-        response.json["response"]["errors"]["setup"][0] == "Marked method is not valid"
+        response.json["response"]["field_errors"]["setup"][0]
+        == "Marked method is not valid"
     )
 
     data = dict(setup="email")
@@ -359,29 +363,29 @@ def test_two_factor_flag(app, client):
     code = sms_sender.messages[0].split()[-1]
     # submit bad token to two_factor_token_validation
     response = client.post("/tf-validate", data=dict(code=wrong_code))
-    assert b"Invalid Token" in response.data
+    assert get_message("TWO_FACTOR_INVALID_TOKEN") in response.data
 
-    # sumbit right token and show appropriate response
+    # submit right token and show appropriate response
     response = client.post("/tf-validate", data=dict(code=code), follow_redirects=True)
-    assert b"Your token has been confirmed" in response.data
+    assert get_message("TWO_FACTOR_LOGIN_SUCCESSFUL") in response.data
 
-    # Upon completion, session cookie shouldnt have any two factor stuff in it.
+    # Upon completion, session cookie shouldn't have any two factor stuff in it.
     assert not tf_in_session(get_session(response))
 
-    # Test change two_factor view to from sms to mail
-    with app.mail.record_messages() as outbox:
-        setup_data = dict(setup="email")
-        response = client.post("/tf-setup", data=setup_data, follow_redirects=True)
-        msg = b"To complete logging in, please enter the code sent to your mail"
-        assert msg in response.data
+    # Test change two_factor view from sms to mail
+    setup_data = dict(setup="email")
+    response = client.post("/tf-setup", data=setup_data, follow_redirects=True)
+    msg = b"To complete logging in, please enter the code sent to your mail"
+    assert msg in response.data
 
-        # Fetch token validate form
-        response = client.get("/tf-validate")
-        assert response.status_code == 200
-        assert b'name="code"' in response.data
+    # Fetch token validate form
+    response = client.get("/tf-validate")
+    assert response.status_code == 200
+    # make sure two_factor_verify_code_form is set
+    assert b'name="code"' in response.data
 
-    code = outbox[0].body.split()[-1]
-    # sumbit right token and show appropriate response
+    code = app.mail.outbox[1].body.split()[-1]
+    # submit right token and show appropriate response
     response = client.post("/tf-validate", data=dict(code=code), follow_redirects=True)
     assert b"You successfully changed your two-factor method" in response.data
 
@@ -419,7 +423,7 @@ def test_two_factor_flag(app, client):
     # Generate token from passed totp_secret
     code = totp.generate().token
     response = client.post("/tf-validate", data=dict(code=code), follow_redirects=True)
-    assert b"Your token has been confirmed" in response.data
+    assert get_message("TWO_FACTOR_LOGIN_SUCCESSFUL") in response.data
 
     # Verify that the remember token is properly set
     found = False
@@ -448,7 +452,7 @@ def test_two_factor_flag(app, client):
     assert b"Open an authenticator app on your device" in response.data
     assert b"data:image/svg+xml;base64," in response.data
 
-    # check appearence of setup page when sms picked and phone number entered
+    # check appearance of setup page when sms picked and phone number entered
     sms_sender = SmsSenderFactory.createSender("test")
     data = dict(setup="sms", phone="+442083661177")
     response = client.post("/tf-setup", data=data, follow_redirects=True)
@@ -457,7 +461,7 @@ def test_two_factor_flag(app, client):
     code = sms_sender.messages[0].split()[-1]
 
     response = client.post("/tf-validate", data=dict(code=code), follow_redirects=True)
-    assert b"Your token has been confirmed" in response.data
+    assert get_message("TWO_FACTOR_LOGIN_SUCCESSFUL") in response.data
     assert not tf_in_session(get_session(response))
 
     logout(client)
@@ -475,18 +479,41 @@ def test_two_factor_flag(app, client):
     data = dict(email="gal2@lp.com", password="password")
     response = client.post("/login", data=data, follow_redirects=True)
     assert b"Please enter your authentication code" in response.data
-    rescue_data = dict(help_setup="lost_device")
+    rescue_data = dict(help_setup="email")
     response = client.post("/tf-rescue", data=rescue_data, follow_redirects=True)
     message = b"The code for authentication was sent to your email address"
     assert message in response.data
-    rescue_data = dict(help_setup="no_mail_access")
+    rescue_data = dict(help_setup="help")
     response = client.post("/tf-rescue", data=rescue_data, follow_redirects=True)
     message = b"A mail was sent to us in order to reset your application account"
     assert message in response.data
 
 
+@pytest.mark.settings(two_factor_rescue_email=False)
+def test_no_rescue_email(app, client):
+    headers = {"Accept": "application/json", "Content-Type": "application/json"}
+
+    response = client.post(
+        "/login", json=dict(email="gal2@lp.com", password="password")
+    )
+    assert response.json["response"]["tf_required"]
+
+    response = client.get("/tf-rescue", headers=headers)
+    options = response.json["response"]["recovery_options"]
+    assert len(options.keys()) == 1
+    assert "help" in options.keys()
+
+    # make sure that even if post using email - we get an error
+    response = client.post("/tf-rescue", json=dict(help_setup="email"))
+    assert response.status_code == 400
+    assert (
+        response.json["response"]["field_errors"]["help_setup"][0]
+        == "Not a valid choice."
+    )
+
+
 @pytest.mark.settings(two_factor_required=True)
-def test_setup_bad_phone(app, client):
+def test_setup_bad_phone(app, client, get_message):
     data = dict(email="matt@lp.com", password="password")
     response = client.post("/login", data=data, follow_redirects=True)
     message = b"Two-factor authentication adds an extra layer of security"
@@ -508,7 +535,7 @@ def test_setup_bad_phone(app, client):
     assert b"data:image/png;base64," not in response.data
 
     response = client.post("/tf-validate", data=dict(code=code), follow_redirects=True)
-    assert b"Your token has been confirmed" in response.data
+    assert get_message("TWO_FACTOR_LOGIN_SUCCESSFUL") in response.data
     assert not tf_in_session(get_session(response))
 
     headers = {"Accept": "application/json", "Content-Type": "application/json"}
@@ -537,7 +564,7 @@ def test_json(app, client):
     # Verify SMS sent
     assert sms_sender.get_count() == 1
     code = sms_sender.messages[0].split()[-1]
-    response = client.post("/tf-validate", json=dict(code=code), headers=headers)
+    response = client.post("/tf-validate", json=dict(code=code))
     assert response.status_code == 200
     # verify logged in
     response = client.get("/profile", follow_redirects=False)
@@ -546,7 +573,7 @@ def test_json(app, client):
 
     # Test that user not yet setup for 2FA gets correct response.
     data = dict(email="matt@lp.com", password="password")
-    response = client.post("/login", json=data, headers=headers)
+    response = client.post("/login", json=data)
     assert response.json["response"]["tf_required"]
     assert response.json["response"]["tf_state"] == "setup_from_login"
 
@@ -557,7 +584,7 @@ def test_json(app, client):
 
     # Now try to setup
     data = dict(setup="sms", phone="+442083661177")
-    response = client.post("/tf-setup", json=data, headers=headers)
+    response = client.post("/tf-setup", json=data)
     assert response.status_code == 200
     assert response.json["response"]["tf_state"] == "validating_profile"
     assert response.json["response"]["tf_primary_method"] == "sms"
@@ -571,11 +598,25 @@ def test_json(app, client):
 
     # Verify tf is now setup and can directly get code
     data = dict(email="matt@lp.com", password="password")
-    response = client.post("/login", json=data, headers=headers)
+    response = client.post("/login", json=data)
     assert response.json["response"]["tf_required"]
     assert response.json["response"]["tf_state"] == "ready"
+
+    # send bad code
+    response = client.post("/tf-validate", json=dict(code="whatsup"))
+    assert response.status_code == 400
+    assert response.json["response"]["field_errors"]["code"][0] == "Invalid code"
+    assert response.json["response"]["errors"][0] == "Invalid code"
+
+    # Do a GET - should get recovery options
+    response = client.get("/tf-validate", headers=headers)
+    options = response.json["response"]["recovery_options"]
+    assert "email" in options.keys()
+    assert "/tf-rescue" in options["email"]
+
+    # now send correct code
     code = sms_sender.messages[0].split()[-1]
-    response = client.post("/tf-validate", json=dict(code=code), headers=headers)
+    response = client.post("/tf-validate", json=dict(code=code))
     assert response.status_code == 200
     # verify logged in
     response = client.get("/profile", follow_redirects=False)
@@ -593,52 +634,47 @@ def test_json(app, client):
 
 @pytest.mark.settings(two_factor_rescue_mail="helpme@myapp.com")
 def test_rescue_json(app, client):
-    headers = {"Accept": "application/json", "Content-Type": "application/json"}
-
-    # it's an error if not logged in.
-    rescue_data_json = dict(help_setup="lost_device")
+    # it's an error if not primary authenticated
+    rescue_data_json = dict(help_setup="help")
     response = client.post(
         "/tf-rescue",
         json=rescue_data_json,
-        headers=headers,
     )
     assert response.status_code == 400
 
     # check when two_factor_rescue function should appear
     data = dict(email="gal2@lp.com", password="password")
-    response = client.post("/login", json=data, headers=headers)
+    response = client.post("/login", json=data)
     assert response.json["response"]["tf_required"]
 
-    with app.mail.record_messages() as outbox:
-        rescue_data = dict(help_setup="lost_device")
-        response = client.post("/tf-rescue", json=rescue_data, headers=headers)
-        assert response.status_code == 200
+    rescue_data = dict(help_setup="email")
+    response = client.post("/tf-rescue", json=rescue_data)
+    assert response.status_code == 200
+    outbox = app.mail.outbox
 
-    assert outbox[0].recipients == ["gal2@lp.com"]
-    assert outbox[0].sender == "no-reply@localhost"
+    assert outbox[0].to == ["gal2@lp.com"]
+    assert outbox[0].from_email == "no-reply@localhost"
     assert outbox[0].subject == "Two-factor Login"
     matcher = re.match(r".*code: ([0-9]+).*", outbox[0].body, re.IGNORECASE | re.DOTALL)
-    response = client.post(
-        "/tf-validate", json=dict(code=matcher.group(1)), headers=headers
-    )
+    response = client.post("/tf-validate", json=dict(code=matcher.group(1)))
     assert response.status_code == 200
     logout(client)
 
     # Try rescue with no email (should send email to admin)
-    client.post("/login", json=data, headers=headers)
-    with app.mail.record_messages() as outbox:
-        rescue_data = dict(help_setup="no_mail_access")
-        response = client.post("/tf-rescue", json=rescue_data, headers=headers)
-        assert response.status_code == 200
+    client.post("/login", json=data)
+    rescue_data = dict(help_setup="help")
+    response = client.post("/tf-rescue", json=rescue_data)
+    assert response.status_code == 200
+    outbox = app.mail.outbox
 
-    assert outbox[0].recipients == ["helpme@myapp.com"]
-    assert outbox[0].sender == "no-reply@localhost"
-    assert outbox[0].subject == "Two-factor Rescue"
-    assert "gal2@lp.com" in outbox[0].body
+    assert outbox[1].to == ["helpme@myapp.com"]
+    assert outbox[1].from_email == "no-reply@localhost"
+    assert outbox[1].subject == "Two-factor Rescue"
+    assert "gal2@lp.com" in outbox[1].body
 
 
 @pytest.mark.settings(two_factor_required=True)
-def test_no_opt_out(app, client):
+def test_no_opt_out(app, client, get_message):
     # Test if 2FA required, can't opt-out.
     sms_sender = SmsSenderFactory.createSender("test")
     response = client.post(
@@ -651,7 +687,7 @@ def test_no_opt_out(app, client):
 
     # submit right token and show appropriate response
     response = client.post("/tf-validate", data=dict(code=code), follow_redirects=True)
-    assert b"Your token has been confirmed" in response.data
+    assert get_message("TWO_FACTOR_LOGIN_SUCCESSFUL") in response.data
 
     response = client.get("/tf-setup", follow_redirects=True)
     assert b"Disable two factor" not in response.data
@@ -700,7 +736,7 @@ def test_evil_validate(app, client):
     del signalled_identity[:]
 
 
-def test_opt_in(app, client):
+def test_opt_in(app, client, get_message):
     """
     Test entire lifecycle of user not having 2FA - setting it up, then deciding
     to turn it back off
@@ -752,7 +788,7 @@ def test_opt_in(app, client):
     assert sms_sender.get_count() == 1
     code = sms_sender.messages[0].split()[-1]
     response = client.post("/tf-validate", data=dict(code=code), follow_redirects=True)
-    assert b"Your token has been confirmed" in response.data
+    assert get_message("TWO_FACTOR_LOGIN_SUCCESSFUL") in response.data
     # Verify now logged in
     with app.app_context():
         user = app.security.datastore.find_user(email="jill@lp.com")
@@ -828,7 +864,7 @@ def test_recoverable(app, client, get_message):
     assert sms_sender.get_count() == 1
     code = sms_sender.messages[0].split()[-1]
     response = client.post("/tf-validate", data=dict(code=code), follow_redirects=True)
-    assert b"Your token has been confirmed" in response.data
+    assert get_message("TWO_FACTOR_LOGIN_SUCCESSFUL") in response.data
 
     # verify we are logged in
     response = client.get("/profile", follow_redirects=False)
@@ -867,7 +903,7 @@ def test_admin_setup_reset(app, client, get_message):
     assert sms_sender.get_count() == 1
     code = sms_sender.messages[0].split()[-1]
     response = client.post("/tf-validate", data=dict(code=code), follow_redirects=True)
-    assert b"Your token has been confirmed" in response.data
+    assert get_message("TWO_FACTOR_LOGIN_SUCCESSFUL") in response.data
 
     # verify we are logged in
     response = client.get("/profile", follow_redirects=False)
@@ -895,7 +931,7 @@ def test_admin_setup_reset(app, client, get_message):
 
 
 @pytest.mark.settings(two_factor_required=True)
-def test_datastore(app, client):
+def test_datastore(app, client, get_message):
     # Test that user record is properly set after proper 2FA setup.
     sms_sender = SmsSenderFactory.createSender("test")
     data = dict(email="gene@lp.com", password="password")
@@ -921,7 +957,7 @@ def test_datastore(app, client):
 
     # submit token and show appropriate response
     response = client.post("/tf-validate", data=dict(code=code), follow_redirects=True)
-    assert b"Your token has been confirmed" in response.data
+    assert get_message("TWO_FACTOR_LOGIN_SUCCESSFUL") in response.data
     session = get_session(response)
     # Verify that successful login clears session info
     assert not tf_in_session(session)
@@ -1060,16 +1096,14 @@ def test_authr_identity_num(app, client):
 )
 def test_email_salutation(app, client):
     authenticate(client, email="jill@lp.com")
-    with app.mail.record_messages() as outbox:
-        response = client.post(
-            "/tf-setup", data=dict(setup="email"), follow_redirects=True
-        )
-        msg = b"To complete logging in, please enter the code sent to your mail"
-        assert msg in response.data
+    response = client.post("/tf-setup", data=dict(setup="email"), follow_redirects=True)
+    msg = b"To complete logging in, please enter the code sent to your mail"
+    assert msg in response.data
+    outbox = app.mail.outbox
 
-    assert "jill@lp.com" in outbox[0].send_to
+    assert "jill@lp.com" in outbox[0].to
     assert "jill@lp.com" in outbox[0].body
-    assert "jill@lp.com" in outbox[0].html
+    assert "jill@lp.com" in outbox[0].alternatives[0][0]
 
 
 @pytest.mark.settings(
@@ -1080,16 +1114,14 @@ def test_email_salutation(app, client):
 )
 def test_username_salutation(app, client):
     authenticate(client, email="jill@lp.com")
-    with app.mail.record_messages() as outbox:
-        response = client.post(
-            "/tf-setup", data=dict(setup="email"), follow_redirects=True
-        )
-        msg = b"To complete logging in, please enter the code sent to your mail"
-        assert msg in response.data
+    response = client.post("/tf-setup", data=dict(setup="email"), follow_redirects=True)
+    msg = b"To complete logging in, please enter the code sent to your mail"
+    assert msg in response.data
+    outbox = app.mail.outbox
 
-    assert "jill@lp.com" in outbox[0].send_to
+    assert "jill@lp.com" in outbox[0].to
     assert "jill@lp.com" not in outbox[0].body
-    assert "jill@lp.com" not in outbox[0].html
+    assert "jill@lp.com" not in outbox[0].alternatives[0][0]
     assert "jill" in outbox[0].body
 
 
@@ -1111,7 +1143,7 @@ def test_bad_sender(app, client, get_message):
     data = dict(email="gal@lp.com", password="password")
     response = client.post("login", json=data, headers=headers)
     assert response.status_code == 500
-    assert response.json["response"]["error"].encode("utf-8") == get_message(
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
         "FAILED_TO_SEND_CODE"
     )
 
@@ -1123,12 +1155,11 @@ def test_bad_sender(app, client, get_message):
 
     response = client.post("tf-setup", json=data, headers=headers)
     assert response.status_code == 500
-    assert response.json["response"]["errors"]["setup"][0].encode(
-        "utf-8"
-    ) == get_message("FAILED_TO_SEND_CODE")
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
+        "FAILED_TO_SEND_CODE"
+    )
 
 
-@pytest.mark.registerable()
 def test_replace_send_code(app, get_message):
     # replace tf_send_code - and have it return an error to check that.
     from flask_sqlalchemy import SQLAlchemy
@@ -1169,7 +1200,7 @@ def test_replace_send_code(app, get_message):
         data = dict(email="trp@lp.com", password="password")
         response = client.post("/login", data=data, follow_redirects=True)
         assert b"Please enter your authentication code" in response.data
-        rescue_data = dict(help_setup="lost_device")
+        rescue_data = dict(help_setup="email")
         response = client.post("/tf-rescue", data=rescue_data, follow_redirects=True)
         assert b"That didnt work out as we planned" in response.data
 
@@ -1177,7 +1208,9 @@ def test_replace_send_code(app, get_message):
         headers = {"Accept": "application/json", "Content-Type": "application/json"}
         response = client.post("/tf-rescue", json=rescue_data, headers=headers)
         assert response.status_code == 500
-        assert response.json["response"]["errors"]["help_setup"][0] == "Failed Again"
+        assert (
+            response.json["response"]["field_errors"]["help_setup"][0] == "Failed Again"
+        )
 
 
 @pytest.mark.settings(freshness=timedelta(minutes=0))
@@ -1236,9 +1269,9 @@ def test_verify_json(app, client, get_message):
 
     response = client.post("verify", json=dict(password="notmine"), headers=headers)
     assert response.status_code == 400
-    assert response.json["response"]["errors"]["password"][0].encode(
-        "utf-8"
-    ) == get_message("INVALID_PASSWORD")
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
+        "INVALID_PASSWORD"
+    )
 
     response = client.post("verify", json=dict(password="password"), headers=headers)
     assert response.status_code == 200
@@ -1312,16 +1345,202 @@ def test_no_sms(app, get_message):
         data = dict(email="trp@lp.com", password="password")
         client.post("/login", data=data, follow_redirects=True)
 
-        with app.mail.record_messages() as outbox:
-            response = client.post(
-                "/tf-setup", data=dict(setup="email"), follow_redirects=True
-            )
-            msg = b"To complete logging in, please enter the code sent to your mail"
-            assert msg in response.data
+        response = client.post(
+            "/tf-setup", data=dict(setup="email"), follow_redirects=True
+        )
+        msg = b"To complete logging in, please enter the code sent to your mail"
+        assert msg in response.data
 
-        code = outbox[0].body.split()[-1]
-        # sumbit right token and show appropriate response
+        code = app.mail.outbox[0].body.split()[-1]
+        # submit right token and show appropriate response
         response = client.post(
             "/tf-validate", data=dict(code=code), follow_redirects=True
         )
         assert b"You successfully changed your two-factor method" in response.data
+
+
+@pytest.mark.settings(multi_factor_recovery_codes=True)
+def test_rc_json(app, client, get_message):
+    # Test recovery codes
+    # gal has two-factor already setup for 'sms'
+    headers = {"Accept": "application/json", "Content-Type": "application/json"}
+    tf_authenticate(app, client)
+
+    response = client.get("/mf-recovery-codes", headers=headers)
+    assert response.status_code == 200
+    codes = response.json["response"]["recovery_codes"]
+    assert len(codes) == 0
+
+    response = client.post("/mf-recovery-codes", headers=headers)
+    codes = response.json["response"]["recovery_codes"]
+    assert len(codes) == 5
+
+    response = client.get("/mf-recovery-codes", headers=headers)
+    recodes = response.json["response"]["recovery_codes"]
+    assert len(recodes) == 5 and codes[0] == recodes[0]
+
+    response = client.get("/mf-recovery-codes", headers=headers)
+    assert response.status_code == 200
+    assert not hasattr(response.json["response"], "recovery_codes")
+
+    # endpoint is for unauthenticated only
+    response = client.post("/mf-recovery", json=dict(code=codes[0]))
+    assert response.status_code == 400
+    logout(client)
+
+    response = client.post("/login", json=dict(email="gal@lp.com", password="password"))
+    assert response.json["response"]["tf_required"]
+
+    response = client.post("/mf-recovery", json=dict(code=codes[0]))
+    assert response.status_code == 200
+
+    # verify actually logged in
+    response = client.get("/profile", follow_redirects=False)
+    assert response.status_code == 200
+
+    # logout and try again - first code shouldn't work again
+    logout(client)
+    response = client.post("/login", json=dict(email="gal@lp.com", password="password"))
+    assert response.json["response"]["tf_required"]
+
+    response = client.post("/mf-recovery", json=dict(code=codes[0]))
+    assert response.status_code == 400
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
+        "INVALID_RECOVERY_CODE"
+    )
+    response = client.post("/mf-recovery", json=dict(code=codes[1]))
+    assert response.status_code == 200
+
+
+@pytest.mark.settings(multi_factor_recovery_codes=True)
+def test_rc(app, client, get_message):
+    # Test recovery codes
+    # gal has two-factor already setup for 'sms'
+    tf_authenticate(app, client)
+
+    response = client.post("/mf-recovery-codes")
+    rd = response.data.decode("utf-8")
+    codes = re.findall(r"[a-f,\d]{4}-[a-f,\d]{4}-[a-f,\d]{4}", rd)
+    assert len(codes) == 5
+
+    response = client.get("/mf-recovery-codes?show_codes=hi")
+    assert response.status_code == 200
+    assert b"Recovery Codes" in response.data
+    rd = response.data.decode("utf-8")
+    codes = re.findall(r"[a-f,\d]{4}-[a-f,\d]{4}-[a-f,\d]{4}", rd)
+    assert len(codes) == 5
+
+    # endpoint is for unauthenticated only
+    response = client.post(
+        "/mf-recovery", data=dict(code=codes[0]), follow_redirects=False
+    )
+    assert response.status_code == 302
+    logout(client)
+
+    response = client.post("/login", json=dict(email="gal@lp.com", password="password"))
+    assert response.json["response"]["tf_required"]
+
+    response = client.post(
+        "/mf-recovery", data=dict(code=codes[0]), follow_redirects=True
+    )
+    assert response.status_code == 200
+
+    # verify actually logged in
+    response = client.get("/profile", follow_redirects=False)
+    assert response.status_code == 200
+
+    # logout and try again - first code shouldn't work again
+    logout(client)
+    response = client.post("/login", data=dict(email="gal@lp.com", password="password"))
+
+    response = client.post("/mf-recovery", data=dict(code=codes[0]))
+    assert get_message("INVALID_RECOVERY_CODE") in response.data
+    response = client.post(
+        "/mf-recovery", data=dict(code=codes[1]), follow_redirects=True
+    )
+    assert response.status_code == 200
+    # verify actually logged in
+    response = client.get("/profile", follow_redirects=False)
+    assert response.status_code == 200
+
+
+@pytest.mark.settings(multi_factor_recovery_codes=True)
+def test_rc_reset(app, client, get_message):
+    # test that reset_user_access, removes recovery codes
+    headers = {"Accept": "application/json", "Content-Type": "application/json"}
+    tf_authenticate(app, client)
+
+    response = client.post("/mf-recovery-codes", headers=headers)
+    codes = response.json["response"]["recovery_codes"]
+    assert len(codes) == 5
+
+    with app.app_context():
+        user = app.security.datastore.find_user(email="gal@lp.com")
+        app.security.datastore.reset_user_access(user)
+        app.security.datastore.commit()
+
+    client.post("/login", json=dict(email="gal@lp.com", password="password"))
+    response = client.get("/mf-recovery-codes", headers=headers)
+    codes = response.json["response"]["recovery_codes"]
+    assert len(codes) == 0
+
+
+@pytest.mark.settings(multi_factor_recovery_codes=True)
+def test_rc_bad_state(app, client, get_message):
+
+    response = client.post("/mf-recovery", json=dict(code="hi"))
+    assert response.status_code == 400
+    assert response.json["response"]["errors"][0].encode("utf=8") == get_message(
+        "TWO_FACTOR_PERMISSION_DENIED"
+    )
+
+
+@pytest.mark.settings(multi_factor_recovery_codes=True)
+def test_rc_rescue(app, client, get_message):
+    headers = {"Accept": "application/json", "Content-Type": "application/json"}
+    tf_authenticate(app, client)
+
+    response = client.post("/mf-recovery-codes", headers=headers)
+    codes = response.json["response"]["recovery_codes"]
+    assert len(codes) == 5
+    logout(client)
+
+    data = dict(email="gal@lp.com", password="password")
+    response = client.post("/login", json=data)
+    assert response.json["response"]["tf_required"]
+
+    response = client.get("/tf-rescue")
+    assert b"Use previously downloaded recovery code" in response.data
+
+    response = client.get("/tf-rescue", headers=headers)
+    options = response.json["response"]["recovery_options"]
+    assert "recovery_code" in options.keys()
+    assert "/mf-recovery" in options["recovery_code"]
+
+    response = client.post(
+        "/tf-rescue", data=dict(help_setup="recovery_code"), follow_redirects=False
+    )
+    assert "/mf-recovery" in response.location
+
+
+@pytest.mark.settings(multi_factor_recovery_codes=True)
+def test_fresh(app, client):
+    # Make sure fetching recovery codes is protected with a freshness check.
+    headers = {"Accept": "application/json", "Content-Type": "application/json"}
+
+    authenticate(client)
+    response = client.post("/mf-recovery-codes", headers=headers)
+    codes = response.json["response"]["recovery_codes"]
+    assert len(codes) == 5
+
+    reset_fresh(client, app.config["SECURITY_FRESHNESS"])
+    response = client.post("/mf-recovery-codes", headers=headers)
+    assert response.status_code == 401
+    assert response.json["response"]["reauth_required"]
+
+    response = client.post("/verify", json=dict(password="password"))
+    assert response.status_code == 200
+
+    response = client.post("/mf-recovery-codes", headers=headers)
+    codes = response.json["response"]["recovery_codes"]
+    assert len(codes) == 5
